/*
 * This file is part of queso.
 *
 * Copyright © 2020 Canek Peláez Valdés
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * Author:
 *    Canek Peláez Valdés <canek@ciencias.unam.mx>
 */

namespace Queso {

    /**
     * Class for the Quesocamera.
     */
    public class Camera : GLib.Object {

        /**
         * The widget.
         */
        public Gtk.Widget widget { public get; private set; }

        /**
         * Whether the camera is saving.
         */
        public bool saving { public get; private set; }

        /* The video and audio devices. */
        private Devices devices;
        private VideoDevice videodev;
        private VideoDevice new_videodev;
        private AudioDevice audiodev;
        private VideoSetup video_setup;
        private AudioSetup audio_setup;

        /* Pipeline and elements. */
        private Gst.Pipeline pipeline;
        private Gst.Bus bus;
        private Gst.Element vsrc;
        private Gst.Element vfilter;
        private Gst.Element vq1;
        private Gst.Element vc;
        private Gst.Element vc2;
        private Gst.Element x264enc;
        private Gst.Element text;
        private Gst.Element info;
        private Gst.Element gtksink;
        private Gst.Element vq2;
        private Gst.Element vq3;
        private Gst.Element tee;
        private Gst.Element mp4mux;
        private Gst.Element filesink;

        // private Gst.Element asrc;
        // private Gst.Element afilter;
        // private Gst.Element parse;
        // private Gst.Element aq1;
        // private Gst.Element aq2;
        // private Gst.Element dec;
        // private Gst.Element ar;
        // private Gst.Element ac;
        // private Gst.Element aenc;

        /* Temporary filename . */
        private string temporary_fn;

        /* Timestamps. */
        private double start_saving_ts;
        private double last_saving_ts;

        private bool show_info;

        /**
         * Constructs a camera object for a video device.
         */
        public Camera(Devices devices) {
            this.devices = devices;
            devices.device_changed.connect(device_changed);
            videodev = devices.video_device;
            audiodev = devices.audio_device;
            video_setup = videodev.setups.first();
            audio_setup = audiodev.setups.first();

            pipeline = new Gst.Pipeline("queso");

            // asrc = audiodev.create_element("asrc");
            // afilter = Gst.ElementFactory.make("capsfilter", "afilter");
            // int rate = 44100;
            // while (rate > audio_setup.max_rate)
            //     rate /= 2;
            // int channels = audio_setup.max_rate > 1 ? 2 : 1;
            // var fmt = audio_setup.formats.first();
            // var acaps = new Gst.Caps.empty_simple(audio_setup.media_type);
            // acaps.set_simple("format", typeof(string), fmt.format,
            //                  "rate", typeof(int), rate,
            //                  "channels", typeof(int), channels);
            // afilter.set("caps", acaps);
            // aq1 = Gst.ElementFactory.make("queue", "aq1");
            // aq2 = Gst.ElementFactory.make("queue", "aq2");
            // aq2.set("max-size-time", 3 * Gst.SECOND);
            // aq2.set("max-size-bytes", 0);
            // aq2.set("max-size-buffers", 0);
            // aq2.set("leaky", 2);
            // ar = Gst.ElementFactory.make("audiorate", "ar");
            // ac = Gst.ElementFactory.make("audioconvert", "ac");
            // aenc = Gst.ElementFactory.make("avenc_aac", "aenc");
            // parse = Gst.ElementFactory.make("h264parse", "parse");
            // parse.set("config_interval", -1);

            vsrc = videodev.create_element("vsrc");
            vfilter = Gst.ElementFactory.make("capsfilter", "vfilter");
            var fr = video_setup.framerates.first();
            var vcaps = new Gst.Caps.empty_simple(video_setup.media_type);
            vcaps.set_simple("framerate", typeof(Gst.Fraction),
                             fr.numerator, fr.denominator,
                             "width", typeof(int), video_setup.width,
                             "height", typeof(int), video_setup.height);
            vfilter.set("caps", vcaps);
            vq1 = Gst.ElementFactory.make("queue", "vq1");
            vc = Gst.ElementFactory.make("videoconvert", "vc");
            text = Gst.ElementFactory.make("textoverlay", "text");
            text.set("font-desc", "sans 14");
            text.set("halignment", 2);
            text.set("valignment", 2);
            info = Gst.ElementFactory.make("textoverlay", "info");
            info.set("font-desc", "sans 10");
            info.set("halignment", 0);
            info.set("valignment", 2);
            info.set("line-alignment", 0);
            gtksink = Gst.ElementFactory.make("gtksink", "gtksink");

            pipeline.add_many(vsrc, vfilter, vq1, vc, text, info, gtksink);
            vsrc.link_many(vfilter, vq1, vc, text, info, gtksink);

            Gtk.Widget w;
            gtksink.get("widget", out w);
            widget = w;

            bus = pipeline.get_bus();

            bus.add_watch(GLib.Priority.DEFAULT, pipeline_bus_watch);
        }

        /**
         * Cleans the camera and stops the pipeline.
         */
        public void clean() {
            pipeline.send_event(new Gst.Event.eos());
            filesink.set_state(Gst.State.NULL);
            mp4mux.set_state(Gst.State.NULL);
            if (saving) {
                var destination_fn = get_destination();
                move_file(temporary_fn, destination_fn);
                destination_fn = temporary_fn = null;
            }
        }

        /**
         * Toggles saving the camera video.
         */
        public void toggle_save() {
            saving = !saving;
            if (saving) {
                start_saving_ts = last_saving_ts = Timestamp.get_timestamp();
                text.set("text", "•REC 00:00:00");
                text.set("halignment", 2);
                text.set("valignment", 2);
                GLib.Idle.add(update_counter);
                start_saving();
            } else {
                stop_saving();
                text.set("text", "");
            }
        }

        /**
         * Starts the camera.
         */
        public void start() {
            GLib.Idle.add(start_camera);
        }

        /**
         * Stops the camera.
         */
        public void stop() {
            GLib.Idle.add(stop_camera);
        }

        /* Starts the camera asynchronously. */
        private bool start_camera() {
            return change_pipeline_state(Gst.State.PLAYING);
        }

        /* Stops the camera asynchronously. */
        private bool stop_camera() {
            return change_pipeline_state(Gst.State.PAUSED);
        }

        /* Pipeline bus watch. */
        private bool
        pipeline_bus_watch(Gst.Bus bus, Gst.Message message) {
            switch (message.type) {
            case Gst.MessageType.ELEMENT:
                unowned Gst.Structure structure = message.get_structure();
                if (!structure.has_name("GstBinForwarded"))
                    return GLib.Source.CONTINUE;
                Gst.Message forward_message = null;
                structure.get("message", typeof(Gst.Message),
                              out forward_message);
                if (forward_message.type != Gst.MessageType.EOS)
                    return GLib.Source.CONTINUE;
                stderr.printf("Finishing file\n");

                tee.set_state(Gst.State.NULL);
                vq2.set_state(Gst.State.NULL);
                vq3.set_state(Gst.State.NULL);
                vc2.set_state(Gst.State.NULL);
                x264enc.set_state(Gst.State.NULL);
                mp4mux.set_state(Gst.State.NULL);
                filesink.set_state(Gst.State.NULL);

                pipeline.remove_many(tee, vq2, vq3, vc2, x264enc,
                                     mp4mux, filesink);
                var src_pad = vq1.get_static_pad("src");
                var sink_pad = vc.get_static_pad("sink");
                src_pad.link(sink_pad);

                var destination_fn = get_destination();
                move_file(temporary_fn, destination_fn);
                temporary_fn = destination_fn = null;
                return GLib.Source.CONTINUE;
            default:
                return GLib.Source.CONTINUE;
            }
        }

        // /* Update counter idle callback. */
        private bool update_counter() {
            if (!saving) {
                text.set("text", "");
                return GLib.Source.REMOVE;
            }
            Gst.State state = Gst.State.NULL;
            Gst.State pending;
            pipeline.get_state(out state, out pending, 100);
            if (state != Gst.State.PLAYING)
                return GLib.Source.CONTINUE;
            double ts = Timestamp.get_timestamp();
            if (ts - last_saving_ts < 1.0)
                return GLib.Source.CONTINUE;
            last_saving_ts = ts;
            string sts = Timestamp.to_string(ts - start_saving_ts);
            text.set("text", "•REC " + sts);

            return GLib.Source.CONTINUE;
        }

        /* Changes the pipeline state, returning until it's changed. */
        private bool change_pipeline_state(Gst.State new_state) {
            pipeline.set_state(new_state);
            Gst.State state = Gst.State.NULL;
            Gst.State pending;

            Gst.StateChangeReturn r;

            do {
                r = pipeline.get_state(out state, out pending, 100);
                if (r == Gst.StateChangeReturn.FAILURE) {
                    return false;
                }
                Thread.usleep(10);
            } while (state != new_state);

            return true;
        }

        /* Gets the final destination for the video. */
        private string get_destination() {
            var now = new GLib.DateTime.now_local();
            string sts = "%04d-%02d-%02d %02d:%02d:%02d".printf(
                now.get_year(), now.get_month(), now.get_day_of_month(),
                now.get_hour(), now.get_minute(), now.get_second());
            var video_dir = GLib.Environment.get_user_special_dir(
                GLib.UserDirectory.VIDEOS);
            return string.join(GLib.Path.DIR_SEPARATOR_S, video_dir,
                               "Queso-%s.mp4".printf(sts));
        }

        /* Moves a file. */
        private void move_file(string src_path, string dest_path) {
            GLib.File src = GLib.File.new_for_path(src_path);
            GLib.File dest = GLib.File.new_for_path(dest_path);

            try {
                src.move(dest, GLib.FileCopyFlags.NONE, null);
            } catch (GLib.Error e) {
                GLib.warning("Could not move “%s” to “%s”: %s",
                             src_path, dest_path, e.message);
            }
        }

        private void start_saving() {
            var pad = vq1.get_static_pad("src");
            pad.add_probe(Gst.PadProbeType.BLOCK | Gst.PadProbeType.BUFFER,
                          start_saving_block_probe);
        }

        private void stop_saving() {
            stderr.printf("stop_saving\n");
            var pad = vq1.get_static_pad("src");
            pad.add_probe(Gst.PadProbeType.BLOCK | Gst.PadProbeType.BUFFER,
                          stop_saving_block_probe);
        }

        /* Pushes the EOS event. */
        private void*
        push_eos_thread() {
            stderr.printf("push_eos_thread\n");
            var pad = vq3.get_static_pad("src");
            var peer = pad.get_peer();
            pipeline.message_forward = true;
            peer.send_event(new Gst.Event.eos());
            return null;
        }

        /* Saving block probe. */
        private Gst.PadProbeReturn
        stop_saving_block_probe(Gst.Pad pad, Gst.PadProbeInfo info) {
            stderr.printf("stop_saving_block_probe: %s::%s\n",
                          pad.parent.name, pad.name);
            pad.remove_probe(info.id);
            var src = vq1.get_static_pad("src");
            var sink = tee.get_static_pad("sink");
            src.unlink(sink);
            src = vq2.get_static_pad("src");
            sink = vc.get_static_pad("sink");
            src.unlink(sink);
            src = vq1.get_static_pad("src");
            sink = vc.get_static_pad("sink");
            src.link(sink);
            new Thread<void*>("eos-push-thread", push_eos_thread);
            return Gst.PadProbeReturn.DROP;
        }

        /* Change video source block probe. */
        private Gst.PadProbeReturn
        start_saving_block_probe(Gst.Pad pad, Gst.PadProbeInfo info) {
            pad.remove_probe(info.id);
            temporary_fn = random_filename();

            var src_pad = vq1.get_static_pad("src");
            var sink_pad = vc.get_static_pad("sink");

            src_pad.unlink(sink_pad);

            tee = Gst.ElementFactory.make("tee", "tee");
            vq2 = Gst.ElementFactory.make("queue", "vq2");
            vq3 = Gst.ElementFactory.make("queue", "vq3");
            vq3.set("leaky", 2);
            vc2 = Gst.ElementFactory.make("videoconvert", "vc2");
            x264enc = Gst.ElementFactory.make("x264enc", "x264enc");
            mp4mux = Gst.ElementFactory.make("mpegtsmux", "mp4mux");
            filesink = Gst.ElementFactory.make("filesink", "filesink");
            filesink.set("async", false);
            filesink.set("location", temporary_fn);

            pipeline.add_many(tee, vq2, vq3, vc2, x264enc, mp4mux, filesink);
            vq1.link(tee);
            tee.link_many(vq2, vc);
            tee.link_many(vq3, vc2, x264enc, mp4mux, filesink);

            new Thread<void*>("add-wait-keyframe-thread",
                              add_wait_keyframe_probe);

            return Gst.PadProbeReturn.DROP;
        }

        private void device_changed(DeviceType device_type) {
            switch (device_type) {
            case VIDEO:
                video_device_changed();
                break;
            }
        }

        private void video_device_changed() {
            text.set("halignment", 1);
            text.set("valignment", 1);
            text.set("text", "changing video source...");
            new_videodev = devices.video_device;
            var pad = vq1.get_static_pad("src");
            pad.add_probe(Gst.PadProbeType.BLOCK |
                          Gst.PadProbeType.BUFFER,
                          change_video_source_block_probe);
        }

        /* Change video source block probe. */
        private Gst.PadProbeReturn
        change_video_source_block_probe(Gst.Pad pad, Gst.PadProbeInfo info) {
            pad.remove_probe(info.id);

            vsrc.set_state(Gst.State.NULL);
            vfilter.set_state(Gst.State.NULL);
            pipeline.remove_many(vsrc, vfilter);

            videodev = new_videodev;
            new_videodev = null;
            video_setup = videodev.setups.first();
            vsrc = videodev.create_element("vsrc");
            vfilter = Gst.ElementFactory.make("capsfilter", "vfilter");
            var fr = video_setup.framerates.first();
            var vcaps = new Gst.Caps.empty_simple(video_setup.media_type);
            vcaps.set_simple("framerate", typeof(Gst.Fraction),
                             fr.numerator, fr.denominator,
                             "width", typeof(int), video_setup.width,
                             "height", typeof(int), video_setup.height);
            vfilter.set("caps", vcaps);
            pipeline.add_many(vsrc, vfilter);
            vsrc.link_many(vfilter, vq1);
            vsrc.set_state(Gst.State.PLAYING);
            vfilter.set_state(Gst.State.PLAYING);

            new Thread<void*>("add-wait-keyframe-thread",
                              add_wait_keyframe_probe);

            return Gst.PadProbeReturn.DROP;
        }

        private void*
        add_wait_keyframe_probe() {
            var pad = vq1.get_static_pad("src");
            pad.add_probe(Gst.PadProbeType.BLOCK | Gst.PadProbeType.BUFFER,
                          wait_keyframe_probe);
            return null;
        }

        private Gst.PadProbeReturn
        wait_keyframe_probe(Gst.Pad pad, Gst.PadProbeInfo info) {
            Gst.Buffer buffer = (Gst.Buffer)info.data;
            var flags = buffer.get_flags();
            if ((flags & Gst.BufferFlags.DELTA_UNIT) == 0) {
                start_saving_ts = last_saving_ts = Timestamp.get_timestamp();
                text.set("text", "");
                return Gst.PadProbeReturn.REMOVE;
            } else {
                return Gst.PadProbeReturn.DROP;
            }
        }

        private int debug_counter;

        public void debug() {
            var output = "queso-pipeline-%03d".printf(debug_counter++);
            stderr.printf("Debug pipeline: %s.dot\n", output);
            Gst.Debug.bin_to_dot_file(pipeline, Gst.DebugGraphDetails.ALL,
                                      output);
        }

        /* Gets a random filename. */
        private string random_filename() {
            int ts = (int)Timestamp.get_timestamp();
            return string.join(GLib.Path.DIR_SEPARATOR_S,
                               GLib.Environment.get_tmp_dir(),
                               "queso-%08X.mp4".printf(ts));
        }

        private string get_info() {
            var r = new StringBuilder();
            r.append("Video source: ");
            r.append(videodev.name);
            r.append("\n");
            r.append("Audio source: ");
            r.append(audiodev.name);
            return r.str;
        }

        public void toggle_info() {
            show_info = !show_info;
            if (show_info) {
                info.set("text", get_info());
            } else {
                info.set("text", "");
            }
        }
    }
}
