/*
 * This file is part of queso.
 *
 * Copyright © 2020 Canek Peláez Valdés
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * Author:
 *    Canek Peláez Valdés <canek@ciencias.unam.mx>
 */

namespace Queso {

    /**
     * Pipeline for video/x-raw.
     *
     * Initial pipeline:
     *
     * ┌────────────────────┐
     * |     video_src      |
     * └─────────┬──────────┘
     *           │
     * ┌─────────┴──────────┐
     * |    video_filter    |
     * └─────────┬──────────┘
     *           │
     * ┌─────────┴──────────┐
     * | video_queue_pretee |
     * └─────────┬──────────┘
     *           │
     * ┌─────────┴──────────┐
     * |   video_convert    |
     * └─────────┬──────────┘
     *           │
     * ┌─────────┴──────────┐
     * | video_text_record  |
     * └─────────┬──────────┘
     *           │
     * ┌─────────┴──────────┐
     * |  video_text_info   |
     * └─────────┬──────────┘
     *           │
     * ┌─────────┴──────────┐
     * |   video_gtksink    |
     * └────────────────────┘
     *
     * Recording pipeline:
     *
     *              ┌────────────────────┐               ┌────────────────────┐
     *              |     video_src      |               |     audio_src      |
     *              └─────────┬──────────┘               └─────────┬──────────┘
     *                        │                                    │
     *              ┌─────────┴──────────┐               ┌─────────┴──────────┐
     *              |    video_filter    |               │    audio_filter    │
     *              └─────────┬──────────┘               └─────────┬──────────┘
     *                        │                                    │
     *              ┌─────────┴──────────┐               ┌─────────┴──────────┐
     *              | video_queue_pretee |               │  audio_queue_rate  │
     *              └─────────┬──────────┘               └─────────┬──────────┘
     *                        │                                    |
     *              ┌─────────┴──────────┐               ┌─────────┴──────────┐
     *              |     video_tee      │               |    audio_rate      │
     *              └─────────┬──────────┘               └─────────┬──────────┘
     *           ┌────────────┴───────────┐                        │
     * ┌─────────┴──────────┐   ┌─────────┴──────────┐   ┌─────────┴──────────┐
     * | video_queue_screen │   │  video_queue_file  │   │   audio_convert    │
     * └─────────┬──────────┘   └─────────┬──────────┘   └─────────┬──────────┘
     *           │                        │                        │
     * ┌─────────┴──────────┐   ┌─────────┴──────────┐   ┌─────────┴──────────┐
     * |   video_convert    |   | video_convert_file |   │   audio_encoder    │
     * └─────────┬──────────┘   └─────────┬──────────┘   └─────────┬──────────┘
     *           │                        │                        │
     * ┌─────────┴──────────┐   ┌─────────┴──────────┐   ┌─────────┴──────────┐
     * | video_text_record  |   │   video_x264enc    │   │ audio_queue_mp4mux │
     * └─────────┬──────────┘   └─────────┬──────────┘   └─────────┬──────────┘
     *           │                        └────────────┬───────────┘
     * ┌─────────┴──────────┐                ┌─────────┴──────────┐
     * |  video_text_info   |                |  mixed_queue_sink  |
     * └─────────┬──────────┘                └─────────┬──────────┘
     *           │                                     │
     * ┌─────────┴──────────┐                ┌─────────┴──────────┐
     * |   video_gtksink    |                |      filesink      │
     * └────────────────────┘                └────────────────────┘
     */
    public abstract class Pipeline : GLib.Object {

        /* The video and audio devices. */
        public VideoDevice videodev;
        public AudioDevice audiodev;
        public VideoSetup video_setup;
        public AudioSetup audio_setup;

        /* Pipeline and elements. */
        protected Gst.Pipeline pipeline;
        public Gst.Bus bus;

        /* Permanent pipeline. */
        public Gst.Element video_src;
        public Gst.Element video_filter;
        public Gst.Element video_queue_pretee;
        public Gst.Element video_convert;
        public Gst.Element video_text_record;
        public Gst.Element video_text_info;
        public Gst.Element video_gtksink;

        /* Audio pipeline. */
        public Gst.Element audio_src;
        public Gst.Element audio_filter;
        public Gst.Element audio_queue_rate;
        public Gst.Element audio_rate;
        public Gst.Element audio_convert;
        public Gst.Element audio_encoder;
        public Gst.Element audio_queue_mp4mux;

        /* Filesink branch. */
        public Gst.Element video_tee;
        public Gst.Element video_queue_screen;
        public Gst.Element video_queue_file;
        public Gst.Element video_convert_file;
        public Gst.Element video_x264enc;
        public Gst.Element video_mp4mux;

        /* Filesink*/
        public Gst.Element mixed_queue_sink;
        public Gst.Element filesink;
    }
}
